#  This file is part of pyresidfp, a SID emulation package for Python.
#
#  Copyright (c) 2018-2023.  Sebastian Klemke <pypi@nerdheim.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import datetime
import logging
import typing as t
from enum import Enum

from ._pyresidfp import ChipModel, SID, SamplingMethod
from .musical_scale import Tone
from .registers import ReadableRegister, WritableRegister


class Voice(Enum):
    """The three voices of an emulated MOS 6581 / MOS 8580"""

    ONE = 0
    TWO = 1
    THREE = 2


class _VoiceRegister(Enum):
    """Voice-specific writable registers"""

    Freq_Lo = WritableRegister.Voice1_Freq_Lo
    Freq_Hi = WritableRegister.Voice1_Freq_Hi
    Pw_Lo = WritableRegister.Voice1_Pw_Lo
    Pw_Hi = WritableRegister.Voice1_Pw_Hi
    Control_Reg = WritableRegister.Voice1_Control_Reg
    Attack_Decay = WritableRegister.Voice1_Attack_Decay
    Sustain_Release = WritableRegister.Voice1_Sustain_Release

    def voice_register(self, voice: Voice) -> WritableRegister:
        """
        Gets the writable register for the given voice.

        Args:
            voice (Voice): Voice to get register for

        Returns:
            WritableRegister according to register type and voice
        """
        return WritableRegister(self.value + 7 * voice.value)


class _Metaclass(type):
    def __init__(
        cls, name: str, bases: t.Tuple[type, ...], attrs: t.Dict[str, t.Any]
    ) -> None:
        super(_Metaclass, cls).__init__(name, bases, attrs)

        for writable_register in WritableRegister:
            setter = type(cls)._create_setter(writable_register)
            setattr(
                cls,
                writable_register.name,
                property(
                    fget=None,
                    fset=setter,
                    doc="int: Writable {0:s} 8-bit register".format(
                        writable_register.name
                    ),
                ),
            )

        for readable_register in ReadableRegister:
            getter = type(cls)._create_getter(readable_register)
            setattr(
                cls,
                readable_register.name,
                property(
                    fget=getter,
                    doc="int: Readable {0:s} 8-bit register".format(
                        readable_register.name
                    ),
                ),
            )

    @staticmethod
    def _create_setter(register: WritableRegister) -> t.Callable[[t.Any, int], None]:
        def setter(self, value: int) -> None:
            self.write_register(register, value)

        return setter

    @staticmethod
    def _create_getter(register: ReadableRegister) -> t.Callable[[t.Any], int]:
        def getter(self) -> int:
            return self.read_register(register)

        return getter


class SoundInterfaceDevice(metaclass=_Metaclass):
    """Emulation for MOS 6581 / MOS 8580 chips."""

    PAL_CLOCK_FREQUENCY: float = 985248.0
    NTSC_CLOCK_FREQUENCY: float = 1022730.0
    DEFAULT_CLOCK_FREQUENCY: float = PAL_CLOCK_FREQUENCY
    DEFAULT_SAMPLING_RATE: float = 48000.0
    DEFAULT_SAMPLING_METHOD: SamplingMethod = SamplingMethod.RESAMPLE
    DEFAULT_CHIP_MODEL: ChipModel = ChipModel.MOS6581

    def __init__(
        self,
        model: t.Optional[ChipModel] = None,
        sampling_method: t.Optional[SamplingMethod] = None,
        clock_frequency: t.Optional[float] = None,
        sampling_frequency: t.Optional[float] = None,
    ) -> None:
        """Creates a new instance."""
        self._log = logging.getLogger(__name__)
        chip_model = model or type(self).DEFAULT_CHIP_MODEL
        sampling_method = sampling_method or type(self).DEFAULT_SAMPLING_METHOD
        clock_frequency = clock_frequency or type(self).DEFAULT_CLOCK_FREQUENCY
        sampling_frequency = sampling_frequency or type(self).DEFAULT_SAMPLING_RATE
        assert 0 < sampling_frequency <= clock_frequency
        assert 125.0 * clock_frequency / sampling_frequency < 16384.0
        self._sid = SID(
            chip_model, sampling_method, clock_frequency, sampling_frequency
        )

    @property
    def chip_model(self) -> ChipModel:
        """ChipModel: Chip model to emulate"""
        return self._sid.chip_model

    @chip_model.setter
    def chip_model(self, value: ChipModel) -> None:
        self._sid.chip_model = value

    @property
    def sampling_method(self) -> SamplingMethod:
        """SamplingMethod: Method to use for sampling"""
        return self._sid.sampling_method

    @sampling_method.setter
    def sampling_method(self, value: SamplingMethod) -> None:
        self._sid.sampling_method = value

    @property
    def clock_frequency(self) -> float:
        """float: System clock frequency"""
        return self._sid.clock_frequency

    @clock_frequency.setter
    def clock_frequency(self, value: float) -> None:
        self._sid.clock_frequency = value

    @property
    def sampling_frequency(self) -> float:
        """float: Sampling frequency"""
        return self._sid.sampling_frequency

    @sampling_frequency.setter
    def sampling_frequency(self, value: float) -> None:
        self._sid.sampling_frequency = value

    def reset(self):
        """Resets the emulation."""
        self._sid.reset()

    def pulse_width(self, voice: Voice, pulse_width: int) -> None:
        """
        Sets the pulse width on a voice.

        Args:
            voice (Voice): Voice to program
            pulse_width (int): 12-bit value to set
        """
        lo, hi = SoundInterfaceDevice._split_pulse_width(pulse_width)

        self._write_voice_register(voice, _VoiceRegister.Pw_Lo, lo)
        self._write_voice_register(voice, _VoiceRegister.Pw_Hi, hi)

    def tone(self, voice: Voice, tone: Tone) -> None:
        """
        Programs the oscillator of a voice to emit a tone from the musical scale.

        Args:
            voice (Voice): Voice to program
            tone (Tone): Tone to program
        """
        oscillator_value = tone.value

        self.oscillator(voice, oscillator_value)

    def oscillator(self, voice: Voice, value: int) -> None:
        """
        Sets the oscillator of a voice to a given 16-bit value.

        Args:
            voice (Voice): Voice to program
            value (int): 16-bit value to program
        """
        lo, hi = SoundInterfaceDevice._split_16_bit(value)

        self._write_voice_register(voice, _VoiceRegister.Freq_Lo, lo)
        self._write_voice_register(voice, _VoiceRegister.Freq_Hi, hi)

    def control(self, voice: Voice, value: int) -> None:
        """
        Sets the control register of a voice.

        Args:
            voice (Voice): Voice to program
            value (int): 8-bit control word to set
        """
        self._write_voice_register(voice, _VoiceRegister.Control_Reg, value)

    def attack_decay(self, voice: Voice, value: int) -> None:
        """
        Sets the attack/decay value of a voice.

        Args:
            voice (Voice): Voice to program
            value (int): 8-bit value to set
        """
        self._write_voice_register(voice, _VoiceRegister.Attack_Decay, value)

    def sustain_release(self, voice: Voice, value: int) -> None:
        """
        Sets the sustain/release value of a voice.

        Args:
            voice (Voice): Voice to program
            value (int): 8-bit value to set
        """
        self._write_voice_register(voice, _VoiceRegister.Sustain_Release, value)

    def mute(self, voice: Voice):
        self._sid.mute(voice.value, True)

    def unmute(self, voice: Voice):
        self._sid.mute(voice.value, False)

    def filter_cutoff(self, value: int) -> None:
        """
        Sets the filter cutoff value.

        Args:
            value (int): 11-bit value to set
        """
        lo, hi = SoundInterfaceDevice._split_filter_cutoff(value)

        self.write_register(WritableRegister.Filter_Fc_Lo, lo)
        self.write_register(WritableRegister.Filter_Fc_Hi, hi)

    def clock(self, duration: datetime.timedelta) -> list:
        """
        Advances system clock for the given duration and samples output.

        Args:
            duration (datetime.timedelta): Duration to emulate

        Returns:
            list of int containing the sampled output in -32768 to 32767 range
        """
        num_cycles = int(duration.total_seconds() * self.clock_frequency)
        num_samples = int(duration.total_seconds() * self.sampling_frequency)

        self._log.debug(
            "Clock for %f cycles (%f samples estimated) for an interval of %s",
            num_cycles,
            num_samples,
            duration,
        )

        # native sample format is signed 16-bit integers in host endianness
        result = self._sid.clock(num_cycles)

        self._log.debug("Retrieved %d samples", len(result))

        return result

    def write_register(self, register: WritableRegister, value: int) -> None:
        """
        Writes an 8-bit value to a writable register.

        Args:
            register (WritableRegister): Register to write to
            value (int): 8-bit value to write
        """
        assert 0 <= value <= (2**8 - 1)

        self._log.debug("Writing value %d to register %s", value, register)

        self._sid.write(register, value)

    def read_register(self, register: ReadableRegister) -> int:
        """
        Reads the current register value from a readable register.

        Args:
            register (ReadableRegister): Register to read from

        Returns:
            int value in the 0 to 255 range
        """
        self._log.debug("Reading register %s", register)

        result = self._sid.read(register)

        self._log.debug("Read value %d from register %s", result, register)

        return result

    def _write_voice_register(
        self, voice: Voice, register: _VoiceRegister, value: int
    ) -> None:
        writable_register = register.voice_register(voice)

        self.write_register(writable_register, value)

    @staticmethod
    def _split_filter_cutoff(filter_cutoff: int) -> tuple:
        assert 0 <= filter_cutoff <= (2**11 - 1)

        lo = filter_cutoff & 0x07
        hi = (filter_cutoff >> 3) & 0xFF

        return lo, hi

    @staticmethod
    def _split_pulse_width(pulse_width: int) -> tuple:
        assert 0 <= pulse_width <= (2**12 - 1)

        lo = pulse_width & 0xFF
        hi = (pulse_width >> 8) & 0x0F

        return lo, hi

    @staticmethod
    def _split_16_bit(word: int) -> tuple:
        assert 0 <= word <= (2**16 - 1)

        lo = word & 0xFF
        hi = (word >> 8) & 0xFF

        return lo, hi



"""
        reSIDfp SID (MOS6581 / MOS8580) extension
        -----------------------------------------

        .. currentmodule:: _pyresidfp

        .. autosummary::
           :toctree: _generate

           SID
    """
from __future__ import annotations
import pyresidfp._pyresidfp
import typing

__all__ = ["ChipModel", "SID", "SamplingMethod"]

class ChipModel:
    """
                   Chip models to emulate.


    Members:

      MOS6581 :
                   The original MOS 6581 chip included in C64


      MOS8580 :
                   The revised MOS 8580 chip included in C64C

    """

    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str:
        """
        :type: str
        """
    @property
    def value(self) -> int:
        """
        :type: int
        """
    MOS6581: pyresidfp._pyresidfp.ChipModel  # value = <ChipModel.MOS6581: 1>
    MOS8580: pyresidfp._pyresidfp.ChipModel  # value = <ChipModel.MOS8580: 2>
    __members__: dict  # value = {'MOS6581': <ChipModel.MOS6581: 1>, 'MOS8580': <ChipModel.MOS8580: 2>}
    pass

class SID:
    """
    MOS6581/MOS8580 emulation.
    """

    def __init__(
        self,
        chip_model: ChipModel,
        method: SamplingMethod,
        clock_frequency: float,
        sampling_frequency: float,
    ) -> None:
        """
        Creates a new instance of SID and sets sampling parameters.

        Use a clock freqency of 985248Hz for PAL C64, 1022730Hz for NTSC C64.
        The default end of passband frequency is pass_freq = 0.9*sample_freq/2
        for sample frequencies up to ~ 44.1kHz, and 20kHz for higher sample frequencies.

        For resampling, the ratio between the clock frequency and the sample frequency
        is limited as follows: 125*clock_freq/sample_freq < 16384
        E.g. provided a clock frequency of ~ 1MHz, the sample frequency can not be set
        lower than ~ 8kHz. A lower sample frequency would make the resampling code
        overfill its 16k sample ring buffer.

        Args:
            chip_model (_pyresidfp.ChipModel):  Chip model to emulate
            method (_pyresidfp.SamplingMethod): Sampling method to use
            clock_frequency (float):            System clock frequency at Hz
            sampling_frequency (float):         Desired output sampling rate

        Raises:
            RuntimeError

        Examples:
            Construct an emulated MOS 6581 chip with resampling sample method, PAL clock frequency,
            and a sampling frequency of 48kHz:

            >>> sid = SID(ChipModel.MOS6581, SamplingMethod.RESAMPLE, 985248.0, 48000.0)
        """
    def clock(self, cycles: int) -> typing.List[int]:
        """
        Clock SID forward using chosen output sampling algorithm and sample.

        Note:
            The number of samples is equal to

            samples = samplingFrequency * cycles / clockFrequency

        Args:
            cycles (int): Number of clock cycles to forward

        Returns:
             :obj:`list` of :obj:`int` samples in range -32768 to 32767
        """
    def enable_filter(self, enable: bool) -> None:
        """
        Enable filter emulation.

        Args:
            enable (bool): False to turn off filter emulation
        """
    def input(self, value: int) -> None:
        """
        16-bit input (EXT IN). Write 16-bit sample to audio input. NB! The caller
        is responsible for keeping the value within 16 bits. Note that to mix in
        an external audio signal, the signal should be resampled to 1MHz first to\
        avoid sampling noise.

        Args:
            value (int): Input level to set
        """
    def mute(self, channel: int, enable: bool) -> None:
        """
        SID voice muting.

        Args:
            channel (int): Channel to modify
            enable (bool): enable muting
        """
    def read(self, offset: int) -> int:
        """
        Read registers.

        Reading a write only register returns the last char written to any SID register.
        The individual bits in this value start to fade down towards zero after a few cycles.
        All bits reach zero within approximately $2000 - $4000 cycles.
        It has been claimed that this fading happens in an orderly fashion,
        however sampling of write only registers reveals that this is not the case.

        Note:
            This is not correctly modeled.

            The actual use of write only registers has largely been made
            in the belief that all SID registers are readable.
            To support this belief the read would have to be done immediately
            after a write to the same register (remember that an intermediate write
            to another register would yield that value instead).
            With this in mind we return the last value written to any SID register
            for $2000 cycles without modeling the bit fading.

        Args:
            offset (int): SID register to read

        Returns:
            char: Value read from chip
        """
    def reset(self) -> None:
        """
        Resets chip model, voice registers, filters and sampling method.

        Raises:
            RuntimeError
        """
    def set_filter_6581_curve(self, curve_position: float) -> None:
        """
        Set filter curve parameter for 6581 model.

        Args:
            curve_position (float): 0 .. 1, where 0 sets center frequency high ("light") and 1 sets it low ("dark"), default is 0.5
        """
    def set_filter_8580_curve(self, curve_position: float) -> None:
        """
        Set filter curve parameter for 8580 model.

        Args:
            curve_position (float):
        """
    def write(self, offset: int, value: int) -> None:
        """
        Write registers.

        Args:
            offset (int): Chip register to write to
            value  (char): Value to write
        """
    @property
    def chip_model(self) -> ChipModel:
        """
                       _pyresidfp.ChipModel: Chip model to emulate.


        :type: ChipModel
        """
    @chip_model.setter
    def chip_model(self, arg1: ChipModel) -> None:
        """
        _pyresidfp.ChipModel: Chip model to emulate.
        """
    @property
    def clock_frequency(self) -> float:
        """
                       float: Clock frequency of chip to emulate


        :type: float
        """
    @clock_frequency.setter
    def clock_frequency(self, arg1: float) -> None:
        """
        float: Clock frequency of chip to emulate
        """
    @property
    def sampling_frequency(self) -> float:
        """
                       float: Frequency at which to sample output


        :type: float
        """
    @sampling_frequency.setter
    def sampling_frequency(self, arg1: float) -> None:
        """
        float: Frequency at which to sample output
        """
    @property
    def sampling_method(self) -> SamplingMethod:
        """
                       _pyresidfp.SamplingMethod: Sampling method to use


        :type: SamplingMethod
        """
    @sampling_method.setter
    def sampling_method(self, arg1: SamplingMethod) -> None:
        """
        _pyresidfp.SamplingMethod: Sampling method to use
        """
    pass

class SamplingMethod:
    """
                   Method to sample emulated anologue output.


    Members:

      DECIMATE :
                   Single-pass down-sampling


      RESAMPLE :
                   Two-pass resampling

    """

    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str:
        """
        :type: str
        """
    @property
    def value(self) -> int:
        """
        :type: int
        """
    DECIMATE: pyresidfp._pyresidfp.SamplingMethod  # value = <SamplingMethod.DECIMATE: 1>
    RESAMPLE: pyresidfp._pyresidfp.SamplingMethod  # value = <SamplingMethod.RESAMPLE: 2>
    __members__: dict  # value = {'DECIMATE': <SamplingMethod.DECIMATE: 1>, 'RESAMPLE': <SamplingMethod.RESAMPLE: 2>}
    pass

__version__ = 1



#  This file is part of pyresidfp, a SID emulation package for Python.
#
#  Copyright (c) 2018-2023.  Sebastian Klemke <pypi@nerdheim.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from .musical_scale import Tone
from .registers import (
    AttackDecayBits,
    ControlBits,
    ModeVolBits,
    ReadableRegister,
    ResFiltBits,
    SustainReleaseBits,
    WritableRegister,
)
from .sound_interface_device import SoundInterfaceDevice, Voice, _VoiceRegister
from ._version import version, version_tuple, __version__, __version_tuple__

__all__ = [
    "Tone",
    "AttackDecayBits",
    "ControlBits",
    "ModeVolBits",
    "ReadableRegister",
    "ResFiltBits",
    "SustainReleaseBits",
    "WritableRegister",
    "SoundInterfaceDevice",
    "Voice",
]



#  This file is part of pyresidfp, a SID emulation package for Python.
#
#  Copyright (c) 2018-2023.  Sebastian Klemke <pypi@nerdheim.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from enum import IntEnum, IntFlag


class WritableRegister(IntEnum):
    """Writable registers on the MOS 6581"""

    Voice1_Freq_Lo = 0x00
    Voice1_Freq_Hi = 0x01
    Voice1_Pw_Lo = 0x02
    Voice1_Pw_Hi = 0x03
    Voice1_Control_Reg = 0x04
    Voice1_Attack_Decay = 0x05
    Voice1_Sustain_Release = 0x06

    Voice2_Freq_Lo = 0x07
    Voice2_Freq_Hi = 0x08
    Voice2_Pw_Lo = 0x09
    Voice2_Pw_Hi = 0x0A
    Voice2_Control_Reg = 0x0B
    Voice2_Attack_Decay = 0x0C
    Voice2_Sustain_Release = 0x0D

    Voice3_Freq_Lo = 0x0E
    Voice3_Freq_Hi = 0x0F
    Voice3_Pw_Lo = 0x10
    Voice3_Pw_Hi = 0x11
    Voice3_Control_Reg = 0x12
    Voice3_Attack_Decay = 0x13
    Voice3_Sustain_Release = 0x14

    Filter_Fc_Lo = 0x15
    Filter_Fc_Hi = 0x16
    Filter_Res_Filt = 0x17
    Filter_Mode_Vol = 0x18


class ReadableRegister(IntEnum):
    """Readable registers on the MOS 6581"""

    Misc_Potx = 0x19
    Misc_Poty = 0x1A
    Misc_Osc3_Random = 0x1B
    Misc_Env3 = 0x1C


class ControlBits(IntFlag):
    """Bits in control registers"""

    GATE = 2**0
    SYNC = 2**1
    RING_MOD = 2**2
    TEST = 2**3
    TRIANGLE = 2**4
    SAWTOOTH = 2**5
    PULSE = 2**6
    NOISE = 2**7


class AttackDecayBits(IntFlag):
    """Bits in Attack/Decay registers"""

    DCY0 = 2**0
    DCY1 = 2**1
    DCY2 = 2**2
    DCY3 = 2**3
    ATK0 = 2**4
    ATK1 = 2**5
    ATK2 = 2**6
    ATK3 = 2**7


class SustainReleaseBits(IntFlag):
    """Bits in Sustain/Release registers"""

    RIS0 = 2**0
    RIS1 = 2**1
    RIS2 = 2**2
    RIS3 = 2**3
    STN0 = 2**4
    STN1 = 2**5
    STN2 = 2**6
    STN3 = 2**7


class ResFiltBits(IntFlag):
    """Bits in Resonance/Filter register"""

    Filt1 = 2**0
    Filt2 = 2**1
    Filt3 = 2**2
    FiltEX = 2**3
    RES0 = 2**4
    RES1 = 2**5
    RES2 = 2**6
    RES3 = 2**7


class ModeVolBits(IntFlag):
    """Bits in Mode/Volume register"""

    VOL0 = 2**0
    VOL1 = 2**1
    VOL2 = 2**2
    VOL3 = 2**3
    LP = 2**4
    BP = 2**5
    HP = 2**6
    THREE_OFF = 2**7



from datetime import timedelta

from pyresidfp import SoundInterfaceDevice, Voice, ControlBits, Tone


def test_sample_length():
    """Emulate SID for 0.62 seconds, expect a corresponding PCM sample vector length"""
    # program SID
    sid = SoundInterfaceDevice()
    sid.Filter_Mode_Vol = 15  # Maximum volume
    sid.attack_decay(Voice.ONE, 190)  # 800 ms attack, 15 s decay
    sid.sustain_release(Voice.ONE, 248)  # sustain peak, 300 ms release
    sid.tone(Voice.ONE, Tone.C4)
    sid.control(Voice.ONE, ControlBits.TRIANGLE | ControlBits.GATE)

    # sample attack phase
    attack_phase = timedelta(seconds=0.32)
    raw_samples = sid.clock(attack_phase)

    # reprogram SID for release phase and sample
    release_phase = timedelta(seconds=0.3)
    sid.control(Voice.ONE, ControlBits.TRIANGLE)
    raw_samples.extend(sid.clock(release_phase))

    expected_vector_length = int(
        (attack_phase.total_seconds() + release_phase.total_seconds())
        * sid.sampling_frequency
    )
    error_margin = int(1 / 100 * expected_vector_length)
    assert (
        expected_vector_length - error_margin
        <= len(raw_samples)
        <= expected_vector_length + error_margin
    )
